from . import DataManager
import pandas as pd
import dataclasses
import datetime
import time
from ..wrapper.mysql import RawDatabaseConnector, BasicDatabaseConnector, DerivedDatabaseConnector
from ..struct import AssetWeight, AssetPrice, FundScoreParam, FundScoreHelper
from .score import FundScoreManager
from .data_tables import FundDataTables
from ..view.raw_models import *
from ..view.basic_models import *
from ..view.derived_models import *

 

class FundDataManager(DataManager):

    def __init__(self, start_time=None, end_time=None, fund_score_param: FundScoreParam=None, score_manager: FundScoreManager=None):
        DataManager.__init__(self, 
            start_time or datetime.datetime(2005, 1, 1), 
            end_time or datetime.datetime.now()
        )
        self._fund_score_param = fund_score_param or FundScoreParam(tag_type=1, score_method=1, is_full=1)
        self.dts = FundDataTables()
        self._is_inited = False
        self._score_manager = score_manager or FundScoreManager()
        self.set_score_param(self._fund_score_param)
    
    @property
    def inited(self):
        return self._is_inited

    def init(self, index_list=None, score_pre_calc=True, print_time=False):
        _tm_start = time.time()
        index_list = index_list or list(AssetWeight.__dataclass_fields__.keys())
        
        def fetch_table(view):
            query = quant_session.query(view)
            return pd.read_sql(query.statement, query.session.bind)

        # info is necessary
        with BasicDatabaseConnector().managed_session() as quant_session:
            _tm_basic_start = time.time()
            self.dts.fund_info = fetch_table(FundInfo)
            _tm_basic_fetch_info = time.time()
            _filter_fund_info = self.dts.fund_info[
                    self.dts.fund_info.index_id.isin(index_list)
                    & ((self.dts.fund_info.structure_type == 0) | (self.dts.fund_info.structure_type == 1)) 
                    & (self.dts.fund_info.is_etf != 1)
                ]
            _tm_basic_filter_info = time.time()
            self.dts.fund_list = set(_filter_fund_info.fund_id)
            self.dts.fund_index_map = {cur.fund_id: cur.index_id for idx, cur in _filter_fund_info.iterrows()}
            _tm_basic_prep_struct = time.time()
            self.dts.trading_days = fetch_table(TradingDayList)
            _tm_basic_fetch_days = time.time()
            # index
            self.dts.index_info = fetch_table(IndexInfo)
            _tm_basic_fetch_index_info = time.time()
            _index_query = quant_session.query(IndexPrice.index_id, IndexPrice.datetime, IndexPrice.close).filter(IndexPrice.index_id.in_(index_list))
            self.dts.index_price = pd.read_sql(_index_query.statement, _index_query.session.bind).pivot_table(index='datetime', columns='index_id', values='close').fillna(method= 'ffill')
            _tm_basic_fetch_index_price = time.time()
            # fund nav
            _fund_nav_query = quant_session.query(
                    FundNav.fund_id,
                    FundNav.adjusted_net_value,
                    FundNav.unit_net_value,
                    FundNav.subscribe_status, 
                    FundNav.redeem_status, 
                    FundNav.datetime
                ).filter(
                    FundNav.fund_id.in_(self.dts.fund_list),
                    FundNav.datetime >= self.start_date,
                    FundNav.datetime <= self.end_date,
                )
            _nav_df = pd.read_sql(_fund_nav_query.statement, _fund_nav_query.session.bind)
            self.dts.fund_nav = _nav_df.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value').fillna(method='ffill')
            self.dts.fund_unit_nav = _nav_df.pivot_table(index='datetime', columns='fund_id', values='unit_net_value').fillna(method='ffill')
            _tm_basic_fetch_fund_nav = time.time()
            if print_time:
                print(f'\t[time][basic] fetch fund info : {_tm_basic_fetch_info - _tm_basic_start}')
                print(f'\t[time][basic] filter fund info: {_tm_basic_filter_info - _tm_basic_fetch_info}')
                print(f'\t[time][basic] prep fund struct: {_tm_basic_prep_struct - _tm_basic_filter_info}')
                print(f'\t[time][basic] fetch trade days: {_tm_basic_fetch_days - _tm_basic_prep_struct}')
                print(f'\t[time][basic] fetch index info: {_tm_basic_fetch_index_info - _tm_basic_fetch_days}')
                print(f'\t[time][basic] fetch indexprice: {_tm_basic_fetch_index_price - _tm_basic_fetch_index_info}')
                print(f'\t[time][basic] fetch fund navs : {_tm_basic_fetch_fund_nav - _tm_basic_fetch_index_price}')
        _tm_basic = time.time()

        # get derived data first
        with DerivedDatabaseConnector().managed_session() as derived_session:
            _tm_deriv_start = time.time()
            # fund indicator
            _fund_indicator_query = derived_session.query(
                    FundIndicator.fund_id,
                    FundIndicator.datetime,
                    FundIndicator.alpha,
                    FundIndicator.beta,
                    FundIndicator.fee_rate,
                    FundIndicator.track_err,
                ).filter(
                    FundIndicator.fund_id.in_(self.dts.fund_list),
                    FundIndicator.datetime >= self.start_date,
                    FundIndicator.datetime <= self.end_date,
                )
            self.dts.fund_indicator = pd.read_sql(_fund_indicator_query.statement, _fund_indicator_query.session.bind)
            _tm_deriv_fetch_fund_indicator = time.time()
            self.dts.fund_indicator['index_id'] = self.dts.fund_indicator.fund_id.apply(lambda x: self.dts.fund_index_map[x])
            _tm_deriv_apply_fund_indicator = time.time()
            if print_time:
                print(f'\t[time][deriv] fetch fund indicator: {_tm_deriv_fetch_fund_indicator - _tm_deriv_start}')
                print(f'\t[time][deriv] apply fund indicator: {_tm_deriv_apply_fund_indicator - _tm_deriv_fetch_fund_indicator}')
        
        _tm_derived = time.time()
        # raw index val pct, only used in tactic asset allocation
        with RawDatabaseConnector().managed_session() as quant_session:
            _tm_raw_start = time.time()
            #query = quant_session.query(IndexValPct)
            query = quant_session.query(EMIndexPct) #东财数据新作的raw 指数估值数据 TODO load from derived and gem use ps_pct
            df = pd.read_sql(query.statement, query.session.bind).pivot_table(index='datetime', columns='index_id', values='pe_pct').fillna(method='ffill')
            _tm_raw_fetch_index_val = time.time()
            self.dts.index_pct = df.rename(columns=FundScoreHelper.TACTIC_ASSET_CONVERT)
            _tm_raw_pivot_index_val = time.time()
            if print_time:
                print(f'\t[time][raw] fetch index val: {_tm_raw_fetch_index_val - _tm_raw_start}')
                print(f'\t[time][raw] pivot raw index: {_tm_raw_pivot_index_val - _tm_raw_fetch_index_val}')
        _tm_raw = time.time()

        self.dts.fix_data()
        _tm_fix = time.time()

        self._is_inited = True
        self._score_manager.set_dts(self.dts)
        if score_pre_calc:
            self._score_manager.pre_calculate()
        _tm_finish = time.time()
        print(self.dts)
        if print_time:
            print(f'[time] basic: {_tm_basic - _tm_start}')
            print(f'[time] deriv: {_tm_derived - _tm_basic}')
            print(f'[time] raw_d: {_tm_raw - _tm_derived}')
            print(f'[time] fix_d: {_tm_fix - _tm_raw}')
            print(f'[time] score: {_tm_finish - _tm_fix}')
            print(f'[time] total: {_tm_finish - _tm_start}')

    def set_score_param(self, score_param: FundScoreParam):
        self._score_manager.set_param(score_param)

    def get_index_pcts(self, dt):
        # jch: only pct within 7 days take effect
        INDEX_PCT_EFFECTIVE_DELAY_DAY_NUM = 7
        res = {}
        for index_id in self.dts.index_pct.columns:
            df = self.dts.index_pct[index_id]
            _filtered = df[(df.index <= dt) & (df.index >= dt - datetime.timedelta(days=INDEX_PCT_EFFECTIVE_DELAY_DAY_NUM))]
            if len(_filtered) > 0:
                res[index_id] = _filtered.iloc[-1]
        return res

    def get_fund_score(self, dt, index_id) -> dict: 
        return self._score_manager.get_fund_score(dt, index_id)

    def get_fund_scores(self, dt) -> dict:
        return self._score_manager.get_fund_scores(dt, self.dts.index_list)

    def get_fund_nav(self, dt):
        df = self.dts.fund_nav
        return df.loc[dt].to_dict()

    def get_fund_unit_nav(self, dt):
        df = self.dts.fund_net_nav
        return df.loc[dt].to_dict()

    def get_fund_info(self, fund_id=None):
        '''
        fund_id                                                   000001!0
        wind_id                                                  000001.OF
        transition                                                       0
        order_book_id                                               000001
        desc_name                                                     华夏成长
        start_date                                              2001-12-18
        end_date                                                2040-12-31
        wind_class_1                                                 混合型基金
        wind_class_2                                               偏股混合型基金
        manager_id       王亚伟(20011218-20050412)\r\n田擎(20040227-20051029...
        company_id                                                    华夏基金
        benchmark                                                        0
        full_name                                               华夏成长证券投资基金
        currency                                                       CNY
        base_fund_id                                                     0
        is_structured                                                    0
        is_open                                                          0
        asset_type                                                    null
        update_time                                             2020-03-24
        manage_fee                                                     1.5
        trustee_fee                                                   0.25
        purchase_fee                                                   NaN
        redeem_fee                                                     NaN
        note                                                          None
        track_index                                                   none
        '''
        if fund_id:
            df = self.dts.fund_info
            _filtered = df[df['fund_id']=='000001!0']
            return None if len(_filtered) == 0 else _filtered.iloc[0]
        else:
            return self.dts.fund_info.copy()

    def get_trading_days(self):
        return self.dts.trading_days.copy()

    def get_index_price(self, dt=None):
        if dt:
            return self.dts.index_price.loc[dt]
        else:
            return self.dts.index_price.copy()
    
    def get_index_price_data(self, dt):
        _index_price = self.get_index_price(dt)
        return AssetPrice(**_index_price.to_dict())

def profile(file_name='/tmp/test.txt'):
    import cProfile, pstats, io
    from pstats import SortKey
    pr = cProfile.Profile()
    pr.enable()
    m = FundDataManager('20190101', '20200101', score_manager=FundScoreManager())
    m.init()
    pr.disable()
    s = io.StringIO()
    sortby = SortKey.CUMULATIVE
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats()
    open(file_name, 'w').write(s.getvalue())


def test():
    #m = FundDataManager('20190101', '20200101', score_manager=FundScoreManager())
    m = FundDataManager('20190101', '20200101')
    m.init()
    print(m.get_fund_info(fund_id=None))
    print(m.get_fund_info(fund_id='000001!0'))
    print(m.get_trading_days())
    print(m.get_fund_score(datetime.date(2019,1,3), 'hs300'))

if __name__ == '__main__':
    profile()


import pandas as pd
import numpy as np
import platform
import matplotlib as mpl
import pylab as pl
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as colors
import matplotlib.dates as mdates
from matplotlib.patches import Rectangle
from ....data.manager.score import FundScoreManager
CURRENT_PLATFORM = platform.system()
if CURRENT_PLATFORM == 'Darwin':
    mpl.rcParams['font.family'] = ['Heiti TC']
else:
    mpl.rcParams['font.family'] = ['STKaiti']

class FundPainter(object):

    @staticmethod
    def plot_fund_weights(  fund_weights_history:dict, 
                            fund_cash_history:dict, 
                            fund_marktet_price_history:dict):
        res = []
        for k,v in fund_weights_history.items():
            v = v
            v['date'] = k
            v['cash'] = fund_cash_history[k]  / fund_marktet_price_history[k]
            res.append(v)
        weights_df = pd.DataFrame(res).set_index('date')
        weights_df = weights_df.div(weights_df.sum(axis=1), axis=0)
        weights_df.fillna(0)[1:].plot.area(figsize=(18,9),legend=False,fontsize = 13)
        s = pl.title('fund weights history', fontsize=20)

    @staticmethod
    def plot_fund_mdd_periods(  fund_mv:pd.DataFrame, 
                                fund_weights_history:dict, 
                                fund_nav:pd.DataFrame, 
                                fund_info:pd.DataFrame):
        df = fund_mv
        mdd_part1 = (df.loc[:, 'mv'] / df.loc[:, 'mv'].rolling(10000, min_periods=1).max())
        mdd_date1 = df.loc[:mdd_part1.idxmin(),'mv'].idxmax()
        mdd_date2 = mdd_part1.idxmin()
        date_list = np.array(list(fund_weights_history.keys()))
        date_list = date_list[(date_list >= mdd_date1) & (date_list <= mdd_date2)]
        fund_list = []
        for d in date_list:
            fund_list.extend([k for k,v in fund_weights_history[d].items() if isinstance(v, float) and (round(v,3) > 0)])
        fund_list = list(set(fund_list)) 
        if 'cash' in fund_list:
            fund_list.remove('cash')
        df = fund_nav[fund_list].loc[mdd_date1:mdd_date2,:] 
        fig, ax = plt.subplots(figsize= [18,9])
        df = df/df.iloc[0]
        table_df = fund_info.set_index('fund_id').loc[fund_list,:].reset_index()[['fund_id','desc_name','index_id']].sort_values(['index_id'])
        df_mdd = pd.DataFrame(((1-df.iloc[-1])*100).round(2))
        df_mdd.columns = ['max draw down %']
        table_df = table_df.set_index('fund_id').join(df_mdd).reset_index()
        fund_list = table_df.fund_id.tolist()
        
        for col in df.columns:
            plt.plot(df.index, df[[col]], linewidth=1.0, label = col)
        plt.legend(fontsize=13 ,loc = 'lower left')
        plt.title('fund bt during mdd period  nav of all funds', fontsize=20)
        ax.xaxis.set_ticks_position('top')
        t = plt.table(cellText=table_df.values.tolist(),
          colLabels=table_df.columns,
          colWidths= [0.25,0.45,0.15,0.15],  
          loc='bottom',
          )
        t.auto_set_font_size(False)
        t.auto_set_font_size(False)
        t.set_fontsize(15)
        t.auto_set_column_width('fund_id')
        t.AXESPAD = 0.1
        t.scale(1, 2)
        plt.show()

    @staticmethod
    def plot_fund_score(fund_mv:pd.DataFrame, 
                        fund_weights_history:dict, 
                        trade_history:dict,
                        index_price:pd.DataFrame,
                        asset_weights:dict,
                        fund_info:pd.DataFrame,
                        fund_nav:pd.DataFrame,
                        fund_score:dict,
                        fund_score_raw:dict,
                        fund_indicator:pd.DataFrame,
                        asset=str,
                        is_tuning=bool,
                        ):
        fund_w = fund_weights_history
        end_date = fund_mv.index.tolist()[-1]
        date_list = list(trade_history.keys()) + [end_date]
        fund_asset_df = fund_info[['fund_id','index_id']].set_index('fund_id')
        fund_indicator = fund_indicator.pivot_table(index = ['fund_id','datetime'])
        traded_to_submit_date = {}
        for d in trade_history:
            dic = trade_history[d][0]
            traded_to_submit_date[dic.trade_date] = dic.submit_date
        res = []
        for k,v in asset_weights.items():
            v['date'] = k
            res.append(v)
        weights_df = pd.DataFrame(res).set_index('date')
        weights_df = weights_df.drop(['cash'], axis = 1).dropna()[1:]
        res = []
        for dic, s in zip(weights_df.to_dict('records'), weights_df.sum(axis = 1).values):
            res.append({ k: v/ s for k, v in dic.items()})
        df = pd.DataFrame(res)
        cols = df.columns.tolist()
        name_dic = fund_info[['fund_id','desc_name']].set_index('fund_id')
        b_d = fund_nav.index[0]
        e_d = fund_nav.index[-1]
        bench_df = index_price.loc[b_d:e_d,[asset]]
        bench_df = bench_df/bench_df.iloc[0]
        for i in range(len(date_list) - 1):   
            b_d = date_list[i]
            e_d = date_list[i+1]
            if b_d == e_d:
                break
            bench_df_tmp = bench_df.loc[b_d:e_d,:]
            f_l = [k  for k ,v in fund_w[b_d].items() if isinstance(v, float) and round(v ,3 ) > 0]
            if 'cash' in f_l:
                f_l.remove('cash')
            f_l = [f for f in f_l if fund_asset_df.loc[f,'index_id'] == asset]
            mv_b = bench_df.loc[b_d,asset]
            fund_tmp = fund_nav.loc[b_d:e_d,f_l].copy()
            fund_tmp = fund_tmp/fund_tmp.iloc[0]
            fund_tmp = fund_tmp*mv_b
            res = []
            submit_d = traded_to_submit_date[b_d]
            if len(f_l) < 1:
                continue
            fig, ax = plt.subplots(figsize= [16,12])
            plt.plot(bench_df_tmp.index, bench_df_tmp[asset], label=asset,linewidth=5.0)
            for f in f_l:
                desc = name_dic.loc[f,'desc_name']
                f_i_dict= fund_indicator.loc[f,submit_d].to_dict()
                dic = {
                    'fund_id' : f,
                    'desc_name' : desc,
                    'score' : round(fund_score[submit_d][asset][f],4),
                    'weight': round(fund_w[b_d][f], 4),
                    'score_raw': round(fund_score_raw[submit_d][asset][f],4),
                }
                for s in ['alpha','beta','fee_rate','track_err']:
                    dic[s] = round(f_i_dict[s],4)
                res.append(dic)    
                plt.plot(fund_tmp.index, fund_tmp[f], label=f+'_'+desc,linestyle='--',linewidth=3.0)    
            
            if is_tuning:
                fund_not_select = [f for f, v in fund_score[submit_d][asset].items() if f not in f_l]
                if len(fund_not_select) < 1:
                    break
                fund_tmp = fund_nav.loc[b_d:e_d,fund_not_select].copy()
                fund_tmp = fund_tmp/fund_tmp.iloc[0]
                fund_tmp = fund_tmp*mv_b
                for f in fund_not_select:
                    desc = name_dic.loc[f,'desc_name']
                    f_i_dict= fund_indicator.loc[f,submit_d].to_dict()
                    dic = {
                        'fund_id' : f,
                        'desc_name' : desc,
                        'score' : round(fund_score[submit_d][asset][f],4),
                        'weight': 0,
                        'score_raw': round(fund_score_raw[submit_d][asset][f],4),
                    }
                    for s in ['alpha','beta','fee_rate','track_err']:
                        dic[s] = round(f_i_dict[s],4)
                    res.append(dic)   
                    plt.plot(fund_tmp.index, fund_tmp[f], label=f+'_'+desc,linestyle=':',linewidth=3.0)    

            plt.legend(fontsize=15 ,loc = 'lower right')
            plt.title(f'{asset} {b_d} {e_d}', fontsize=25)
            plt.suptitle(FundScoreManager().funcs[asset].__dict__, y=0.87, fontsize=18)
            ax.xaxis.set_ticks_position('top')
            fund_df = pd.DataFrame(res)
            fund_df = fund_df[['desc_name','fund_id','weight','alpha','beta','track_err','fee_rate','score_raw','score']]
            fund_df = fund_df.sort_values('score', ascending = False)
            t = plt.table(
                cellText=fund_df.values.tolist(),
                colLabels=fund_df.columns,
                loc='bottom',
                colWidths= [0.25,0.12,0.09,0.09,0.09,0.09,0.09,0.09,0.09]          
            )
            t.auto_set_font_size(False)
            t.set_fontsize(15)
            t.auto_set_column_width('fund_id')
            t.AXESPAD = 0.1
            t.scale(1, 4)
            plt.show()



from .config import Configurator
from .struct import WxRawData, WxSeqItem, WxMsgItem
from .processor import IfaProcessor
import time, json

import sys
import libwxwork

class WxworkRunner(object):

    def __init__(self, seq=None):
        self.seq = seq
        corp_id = Configurator().config['wxwork']['corp_id']
        secret_key = Configurator().config['wxwork']['secret_key']
        private_key = Configurator().config['wxwork']['private_key']
        with open(private_key, 'r') as fin:
            private_key_content = fin.read()
        self.wrapper = libwxwork.WxworkWrapper(corp_id, secret_key, private_key_content, -1)
        self.processors = []

    def register(self, processor: IfaProcessor):
        self.processors.append(processor)

    def run(self):
        self.wrapper.init()
        self.seq = self.seq or 0
        while True:
            time.sleep(1)
            res = self.wrapper.get(self.seq, 500)
            try:
                data = WxRawData(**json.loads(res))
                for item in data.chatdata:
                    self.process(item)
            except Exception as e:
                print(e)
            self.seq = data.max_seq or self.seq

    def process(self, item: WxSeqItem):
        decrypted = None
        try:
            decrypted_str = self.wrapper.decode(item.encrypt_random_key, item.encrypt_chat_msg)
            decrypted = json.loads(decrypted_str)
        except:
            print('!' * 10)
            print(f'error: {decrypted_str}')
            print('!' * 10)
        if decrypted:
            msg = WxMsgItem(decrypted)
            for pr in self.processors:
                pr.process(msg)



import math
import pandas as pd
import pickle as pkl
from ..wrapper.mysql import DerivedDatabaseConnector
from ..view.derived_models import *

class DerivedDataApi(object):
    def get_fund_indicator(self, fund_list):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                query = quant_session.query(
                        FundIndicator.fund_id,
                        FundIndicator.datetime,
                        FundIndicator.alpha,
                        FundIndicator.beta,
                        FundIndicator.fee_rate,
                        FundIndicator.track_err,
                    ).filter(
                        FundIndicator.fund_id.in_(fund_list),
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))
        
    def get_fund_score(self):
        with DerivedDatabaseConnector().managed_session() as quant_session:
            try:
                query = quant_session.query(
                        FundScore,   
                    )
                tag_df = pd.read_sql(query.statement, query.session.bind)
                return tag_df
            
            except Exception as e:
                print('Failed to get data <err_msg> {}'.format(e))

    def get_index_valuation(self, index_id, start_date):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    IndexValuation
                ).filter(
                    IndexValuation.index_id == index_id,
                    IndexValuation.datetime >= start_date,
                ).order_by(IndexValuation.datetime.asc())
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(e)
                return pd.DataFrame([])

    def get_index_valuation_develop(self, index_id, start_date):
        with DerivedDatabaseConnector().managed_session() as mn_session:
            try:
                query = mn_session.query(
                    IndexValuationDevelop
                ).filter(
                    IndexValuationDevelop.index_id == index_id,
                    IndexValuationDevelop.datetime >= start_date,
                ).order_by(IndexValuationDevelop.datetime.asc())
                df = pd.read_sql(query.statement, query.session.bind)
                return df
            except Exception as e:
                print(e)
                return pd.DataFrame([])




import pandas as pd
import numpy as np
import datetime
from pyfinance.ols import PandasRollingOLS
import traceback
from ...manager.manager_fund import FundDataManager
from ...manager.score import FundScoreManager
from ...struct import AssetWeight
from ...wrapper.mysql import DerivedDatabaseConnector
from ...view.derived_models import FundIndicator

class FundIndicatorProcessor:

    TRADING_DAYS_PER_YEAR = 242
    SHORT_TIME_SPAN = TRADING_DAYS_PER_YEAR
    LONG_TIME_SPAN = TRADING_DAYS_PER_YEAR * 3
    LONG_TERM_INDEX = ['hs300', 'csi500']

    def init(self, start_date, end_date):    
        self._dm = FundDataManager(start_time=start_date, end_time=end_date, score_manager=FundScoreManager())
        self._dm.init()
        self.start_date = self._dm.start_date
        self.end_date = self._dm.end_date
        self.fund_nav = self._dm.dts.fund_nav.copy()
        self.fund_info = self._dm.dts.fund_info.copy()
        self.index_price = self._dm.dts.index_price.copy()
        
        # 指数日线价格和基金净值 时间轴对齐 ，避免join后因为日期不存在出现价格空值
        # 指数数据历史上出现非交易日有数据，join后，基金净值出现空值，长周期时间序列算指标出空值 
        # 避免对日收益fillna(0)， 造成track error 计算不符合实际
        # 计算逻辑空值填充情况： 1. 只对基金费率空值fillna(0) 
        #                    2. manager fund 里对日线价格ffill
        #                    3. 指数数据按照基金数据重做index后ffill(有基金数据对日子,没有指数数据，比如季报出在季度末，非交易日有基金净值)
        #                    4. 有对超过基金终止日净值赋空值
        end_date = min(self.fund_nav.index[-1], self.index_price.index[-1])
        self.fund_nav = self.fund_nav.loc[:end_date]
        self.index_price = self.index_price[:end_date]
        fund_nav_index = self.fund_nav.index
        self.index_price = self.index_price.reindex(fund_nav_index).fillna(method = 'ffill')
        l1 = self.fund_nav.index.to_list().sort()
        l2 = self.index_price.index.to_list().sort()
        assert l1 == l2, f'date index of index price and fund nav are not identical, l1 {l1} lens {len(l1)}, l2 {l2} lens {len(l2)}'

        fund_to_enddate_dict = self.fund_info[['fund_id', 'end_date']].set_index('fund_id').to_dict()['end_date']
        self.fund_to_index_dict = self.fund_info[['fund_id', 'index_id']].set_index('fund_id').to_dict()['index_id']
        # 超过基金终止日的基金净值赋空
        for fund_id in self.fund_nav.columns:
            fund_end_date = fund_to_enddate_dict[fund_id]
            if self.end_date > fund_end_date:
                self.fund_nav.loc[self.end_date:,fund_id] = np.nan

        self.index_ret = np.log(self.index_price / self.index_price.shift(1))
        self.fund_ret = np.log(self.fund_nav / self.fund_nav.shift(1))
        self.fund_ret = self.fund_ret.stack().reset_index(level=[0,1]).rename(columns={0:'ret'})        
        self.fund_ret['index_id'] = self.fund_ret.fund_id.apply(lambda x: self.fund_to_index_dict[x])
        self.fund_ret = self.fund_ret.pivot_table(index = ['index_id','datetime'],columns='fund_id',values='ret')
        self.index_ret = np.log(self.index_price / self.index_price.shift(1))
        self.fund_to_index_dict = {fund_id:index_id for fund_id,index_id in self.fund_to_index_dict.items() if fund_id in self.fund_ret.columns}
        self.index_list = self.fund_ret.index.levels[0]
        self.index_fund = { index_id : [fund_idx for fund_idx, index_idx in self.fund_to_index_dict.items() if index_idx == index_id] for index_id in self.index_list}
        
    def get_time_range(self, index_id):
        if index_id in self.LONG_TERM_INDEX:
            return self.LONG_TIME_SPAN
        else:
            return self.SHORT_TIME_SPAN

    def get_beta(self, fund_ret, index_ret, time_range):
        # beta 贝塔 
        # 表示基金的收益相对与基准对变化， 绝对值越大， 相对于基准波动越大， 体现为风险越大
        # rf: 基金日收益  rm: 基准日收益 
        # rf = beta * rm + C （滚动计算）
        res = []
        index_id = index_ret.columns[0]
        for fund_id in fund_ret.columns:
            df = fund_ret[[fund_id]].join(index_ret)
            x = df[index_id]
            y = df[fund_id]
            model = PandasRollingOLS(y=y, x=x, window=time_range)
            res_i = model.beta
            res_i.columns = [fund_id]
            res.append(res_i)
        return res

    def get_alpha(self,fund_ret, index_ret, time_range):
        # alpha 阿尔法
        # 表示超越市场的收益
        # rf_avg 基金平均收益 rb_avg 基准平均收益
        # alpha = (rf_avg - beta * rb_avg) * year
        fund_list = fund_ret.columns.tolist()
        index_id = index_ret.columns[0]
        rf_avg = fund_ret.rolling(time_range).mean().reindex(self.beta.index)
        rb_avg = index_ret.rolling(time_range).mean().reindex(self.beta.index)
        v1 = self.beta[fund_list].multiply(rb_avg.values, axis=index_id).copy()
        alpha_i = (rf_avg - v1) * self.TRADING_DAYS_PER_YEAR
        return alpha_i
    
    def get_track_error(self,fund_ret, index_ret, time_range):
        # track error 跟踪误差
        # 表示跟踪基准的能力
        # track_error = (rf-rb).std
        fund_list = fund_ret.columns.tolist()
        index_id = index_ret.columns[0]
        res = []
        for fund_id in fund_list:
            df = fund_ret[[fund_id]].join(index_ret)
            track_err_i = (df[fund_id] - df[index_id]).rolling(time_range).std( ddof=1 ) #按照天天基金的算法， 不年化
            res.append(track_err_i)
        df = pd.DataFrame(res).T
        df.columns = fund_list
        return df

    def upload_derived(self, df, table_name):
        print(table_name)
        df.to_sql(table_name, DerivedDatabaseConnector().get_engine(), index=False, if_exists='append')

    def calculate(self):
        res = []
        for index_id in self.index_list:
            time_range = self.get_time_range(index_id)
            fund_list = self.index_fund[index_id]
            fund_ret = self.fund_ret.loc[index_id][fund_list].copy()
            index_ret = self.index_ret[[index_id]]
            res.extend(self.get_beta(fund_ret, index_ret, time_range))
        self.beta = pd.concat(res, axis=1, sort=True) 
        
        res_alpha = []
        res_track_err = []
        for index_id in self.index_list:
            time_range = self.get_time_range(index_id)
            fund_list = self.index_fund[index_id]
            fund_ret = self.fund_ret.loc[index_id][fund_list].copy()
            index_ret = self.index_ret[[index_id]]
            res_alpha.append(self.get_alpha(fund_ret, index_ret, time_range))
            res_track_err.append(self.get_track_error(fund_ret, index_ret, time_range))
        
        self.alpha = pd.concat(res_alpha, axis=1, sort=True)  
        self.track_err = pd.concat(res_track_err, axis=1, sort=True)  

        fac_list = ['beta','alpha','track_err']
        data_list = [self.beta, self.alpha, self.track_err]
        res = []
        for name, data in zip(fac_list, data_list):
            df = data.stack().reset_index(level=[0,1]).copy()
            df = df.rename(columns = {'level_0':'datetime','level_1':'fund_id',0:'value'})
            df['name'] = name
            res.append(df.copy())
        self.result = pd.concat(res).pivot_table(index=['datetime','fund_id'], columns='name',values='value')

        self.fund_fee = self.fund_info[['fund_id','manage_fee','trustee_fee']].set_index('fund_id').fillna(0).sum(axis = 1)
        self.fund_fee = pd.DataFrame(self.fund_fee).rename(columns = {0:'fee_rate'})
        span_dic = {}
        for index_id, fund_list in self.index_fund.items():
            time_span = int(self.get_time_range(index_id) / self.TRADING_DAYS_PER_YEAR)
            for fund_id in fund_list:
                span_dic[fund_id] = time_span
        self.fund_fee['timespan'] = self.fund_fee.index.map(lambda x : span_dic.get(x,0)).tolist()
        self.result = self.result.join(self.fund_fee).reset_index(level=[0,1])

    def process(self, start_date, end_date):
        failed_tasks = []
        try:
            start_date_dt = datetime.datetime.strptime(start_date, '%Y%m%d').date()
            end_date_dt = datetime.datetime.strptime(end_date, '%Y%m%d').date()

            start_date = start_date_dt - datetime.timedelta(days = 1150) #3年历史保险起见，多取几天 3*365=1095 
            start_date = datetime.datetime.strftime(start_date, '%Y%m%d')

            self.init(start_date, end_date)
            self.calculate()
            df = self.result[(self.result['datetime'] >= start_date_dt) & (self.result['datetime'] <= end_date_dt)]

            self.upload_derived(df, FundIndicator.__table__.name)
        except Exception as e:
            print(e)
            traceback.print_stack()
            failed_tasks.append('fund_indicator')

        return failed_tasks



from ...data.struct import AssetWeight, AssetPrice, AssetPosition, AssetValue
from ...data.struct import AssetTrade, FundTrade, FundPosition
from ...data.struct import AssetTradeParam, FundTradeParam
from . import Helper
import numpy as np
class AssetTrader(Helper):

    def __init__(self, asset_param: AssetTradeParam=None):
        self.asset_param = asset_param or AssetTradeParam()

    def calc_asset_trade(self, dt, 
                               cur_position: AssetPosition, 
                               cur_price: AssetPrice, 
                               target_allocation: AssetWeight):
        cur_mv = AssetValue(prices=cur_price, positions=cur_position)
        tot_mv = cur_mv.sum()

        trades = []
        launch_trade = False
        for index_id, target_weight in target_allocation.__dict__.items():
            if index_id != 'cash':
                target_amt = tot_mv * target_weight
                p = cur_price.__dict__[index_id]
                cur_amt = cur_mv.__dict__[index_id]
                if abs(target_amt - cur_amt) > tot_mv * self.asset_param.MinCountAmtDiff:
                    amount = abs(target_amt - cur_amt)
                    is_buy = target_amt > cur_amt
                    trades.append(AssetTrade(
                        index_id=index_id, 
                        mark_price=p, 
                        amount=amount, 
                        is_buy=is_buy,
                        submit_date=dt
                    ))
                launch_trade = launch_trade or abs(target_amt - cur_amt) > tot_mv * self.asset_param.MinActionAmtDiff

        
        if not launch_trade:
            return cur_position, None
        else:
            trades.sort(key=lambda x: x.is_buy)       
            new_position = cur_position.copy()
            for trd in trades:
                new_position.update(trd)
            return new_position, trades

    def finalize_trade(self, dt, trades: list, t1_price: AssetPrice, bt_position: AssetPosition):
        pendings = []
        traded_list = []
        if trades is None or len(trades) == 0:
            return pendings, traded_list
        # TODO: if some trades needs more time
        for trd in trades:
            # TODO: commision calculate
            #trd.commission = ?
            # update position
            trd.trade_price = t1_price.__dict__[trd.index_id]
            trd.volume = trd.volume if trd.volume else (trd.amount / trd.trade_price)
            trd.trade_date = dt
            if not trd.is_buy:
                if not(bt_position.__dict__[trd.index_id] - trd.volume > -1e-8):
                    print(f'trade volume exceeds, adjusted to pos (index_id){trd.index_id} (vol){trd.volume} (is_buy){trd.is_buy} (pos){bt_position.__dict__[trd.index_id]}')
                    trd.volume = bt_position.__dict__[trd.index_id]
                    trd.amount = trd.volume * trd.trade_price
            bt_position.update(trd)
            traded_list.append(trd)
        return pendings, traded_list

class FundTrader(AssetTrader):

    def __init__(self, asset_param: AssetTradeParam=None, fund_param: FundTradeParam=None):
        AssetTrader.__init__(self, asset_param=asset_param)
        self.fund_param = fund_param or FundTradeParam()

    def calc_fund_trade(self, dt,
                              cur_asset_position: AssetPosition, 
                              cur_asset_price: AssetPrice,
                              target_allocation: AssetWeight, 
                              cur_fund_position: FundPosition, 
                              cur_fund_nav: dict, # fund_id -> nav
                              cur_fund_score: dict): # index_id -> fund_id -> score
        target_asset_position, trade_list = self.calc_asset_trade(dt,
            cur_asset_position, cur_asset_price, target_allocation)
        # 大类资产是否需要调仓的标志
        asset_needs_adjust = trade_list != None and len(trade_list) > 0
        # 如果资产需要调整，那么统一看最优的基金配置是多少（这里如果资产不需要调整，那么基金也不会调）
        if asset_needs_adjust:
            # init
            new_fund_position = cur_fund_position.copy()
            fund_trades = []

            target_asset_mv = AssetValue(prices=cur_asset_price, positions=target_asset_position)
            target_asset_wgts = target_asset_mv.get_weight()
            fund_tot_mv, cur_fund_wgts = cur_fund_position.calc_mv_n_w(fund_navs=cur_fund_nav)
            
            for index_id in target_asset_wgts.__dict__.keys():
                asset_wgt = target_asset_wgts.__dict__[index_id]
                _scores = sorted(cur_fund_score.get(index_id, {}).items(), key=lambda item: item[1], reverse=True)
                fund_weights = {}
                fund_tot_weight = 0
                fund_selected_num = 0
                for i in range(0, len(_scores)):
                    fund_id, score_as_weight = _scores[i]
                    # 最多不超过 MaxFundNumUnderAsset 个基金
                    if not np.isnan(cur_fund_nav[fund_id]):
                        fund_weights[fund_id] = score_as_weight
                        fund_tot_weight += score_as_weight
                        fund_selected_num += 1
                        if fund_selected_num == self.fund_param.MaxFundNumUnderAsset:
                            break
                #assert fund_selected_num > 0, f'index {index_id} must have at least scored fund'
                # normalize
                fund_weights = {fund_id: fund_weights[fund_id] / fund_tot_weight for fund_id in fund_weights.keys()}
                # 计算每只基金的具体仓位变动情况
                for fund_id in set(list(fund_weights.keys()) + cur_fund_position.get_funds(index_id)):
                    target_fund_amt = fund_weights.get(fund_id, 0) * asset_wgt * fund_tot_mv
                    cur_fund_volume = cur_fund_position.get_volume(fund_id) or 0
                    p = cur_fund_nav[fund_id]
                    cur_fund_amt = cur_fund_volume * p

                    if abs(target_fund_amt - cur_fund_amt) > fund_tot_mv * self.fund_param.MinCountAmtDiff or (target_fund_amt == 0 and cur_fund_amt > 0):
                        # TODO: commision and 如果是清某一只基金的逻辑，清空可以执行
                        _trade = FundTrade(
                            fund_id=fund_id,
                            index_id=index_id, 
                            mark_price=p, 
                            amount=abs(target_fund_amt - cur_fund_amt),
                            is_buy=target_fund_amt > cur_fund_amt,
                            submit_date=dt
                        )
                        fund_trades.append(_trade)
                        # print(f'(fund){fund_id} (p){p} (amt0){cur_fund_amt} (amt1){target_fund_amt} (idx){index_id} (tar){asset_mv}')

            fund_trades.sort(key=lambda x: x.is_buy)
            for _trade in fund_trades:
                new_fund_position.update(_trade)
            trade_list += fund_trades
            return new_fund_position, trade_list
        else:
            return cur_fund_position, None


    def finalize_trade(self, dt, trades: list, 
                            t1_price: AssetPrice, 
                            bt_position: AssetPosition,
                            cur_fund_position: FundPosition, 
                            cur_fund_nav: dict):

        if trades is None or len(trades) == 0:
            return [], []
        asset_trades = []
        fund_trades = []
        for trd in trades:
            if isinstance(trd, FundTrade):
                fund_trades.append(trd)
            else:
                asset_trades.append(trd)
        
        pendings, traded_list = AssetTrader.finalize_trade(self, dt, asset_trades, t1_price, bt_position)
        # TODO: if some trades needs more time
        for trd in fund_trades:
            # TODO: commision calculate
            trd.trade_price = cur_fund_nav[trd.fund_id]
            trd.volume = trd.volume if trd.volume else (trd.amount / trd.trade_price)
            trd.trade_date = dt            
            if not trd.is_buy:
                cur_vol = cur_fund_position.get_volume(trd.fund_id)
                if not((cur_vol or 0) - trd.volume > -1e-8):
                    print(f'trade volume exceeds, adjusted to pos (fund_id){trd.fund_id} (vol){trd.volume} (is_buy){trd.is_buy} (pos){cur_vol}')
                    assert cur_vol is not None, 'sell fund with no current position!'
                    trd.volume = cur_vol
                    trd.amount = trd.volume * trd.trade_price
            trade_status = cur_fund_position.update(trd)
            
            assert trade_status, f"trade failed {trd}"
            if trade_status:
                traded_list.append(trd)
        #print('trade: ', trades)
        #print()

        return pendings, traded_list



from .util import Singleton
import platform
import json

CURRENT_PLATFORM = platform.system()
DEFAULT_CONFIG_PATH = '/shared/etc/ifa/config.json'

class MessageSenderSetting(object):

    def __init__(self, data_json):
        assert 'corp_id' in data_json, 'could not find corp_id in config file'
        assert 'app_id' in data_json, 'could not find app_id in config file'
        assert 'app_secret' in data_json, 'could not find add_secret in config file'

        self.corp_id = data_json['corp_id']
        self.app_id = data_json['app_id']
        self.app_secret = data_json['app_secret']

class Configurator(metaclass=Singleton):

    def __init__(self, config_path=DEFAULT_CONFIG_PATH):
        with open(config_path) as f:
            self.config = json.load(f)

    def get_message_sender_setting(self):
        settings = MessageSenderSetting(self.config['message_sender'])
        return settings



import pandas as pd
import numpy as np
import datetime
import copy
from pprint import pprint 
import platform
import matplotlib as mpl
import pylab as pl
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import matplotlib.colors as colors
import matplotlib.dates as mdates
from matplotlib.patches import Rectangle
from .asset_helper import SAAHelper, TAAHelper
from .painter.asset_painter import AssetPainter
from .painter.fund_painter import FundPainter
from .painter.hybrid_painter import HybridPainter
from ...data.manager.manager_fund import FundDataManager
from ...data.struct import AssetWeight, AssetPosition, AssetPrice, AssetValue
from ...data.struct import FundPosition, FundPosItem, FundTrade
from ...data.manager.score import FundScoreManager

CURRENT_PLATFORM = platform.system()
if CURRENT_PLATFORM == 'Darwin':
    mpl.rcParams['font.family'] = ['Heiti TC']
else:
    mpl.rcParams['font.family'] = ['STKaiti']

class ReportHelper:
    
    '''
    save backtest history
    '''
    TRADING_DAYS_PER_YEAR = 242

    def __init__(self):
        pass

    def init(self): 
        pass

    def plot_init(self, dm, taa_helper):
        self.index_price = dm.get_index_price()
        self.fund_info = dm.get_fund_info()
        self.fund_nav = dm.dts.fund_nav
        self.fund_indicator = dm.dts.fund_indicator
        self.taa_params = taa_helper.params if isinstance(taa_helper, TAAHelper) else None
        self.index_pct = dm.dts.index_pct

    def setup(self, saa:AssetWeight):
        self.saa_weight = saa.__dict__
        self.asset_cash_history = {}
        self.asset_position_history = {}
        self.asset_market_price_history = {}
        self.pending_trade_history = {}
        self.asset_weights_history = {}
        self.tactic_history = {}
        self.fund_position_history = {}
        self.trade_history = {}
        self.rebalance_date = []
        self.fund_cash_history = {}
        self.fund_marktet_price_history = {}
        self.fund_weights_history = {}
        self.fund_score = {}
        self.fund_score_raw = {}
        self.target_allocation = {}

    def update(self,dt:datetime, asset_cur_position:AssetPosition, asset_cur_price:AssetPrice, pend_trades:list, fund_cur_position:FundPosition, fund_nav:dict, traded_list:list, fund_score:dict, fund_score_raw:dict, target_allocation):   
        # 检查回测时使用
        self.asset_cash_history[dt] = asset_cur_position.cash##
        self.asset_position_history[dt] = asset_cur_position.__dict__##
        self.pending_trade_history[dt] = pend_trades##
        if fund_cur_position is not None:
            dic = {f : f_info.__dict__  for f, f_info in fund_cur_position.funds.items()}
            for f, f_info in dic.items():
                f_info['price'] =  fund_nav[f]
            self.fund_position_history[dt] = dic##        
            self.fund_cash_history[dt] = fund_cur_position.cash
            mv, fund_w = fund_cur_position.calc_mv_n_w(fund_navs=fund_nav)
            self.fund_marktet_price_history[dt] = mv
            self.fund_weights_history[dt] = { fund_id : w_i for fund_id, w_i in fund_w.items() if w_i > 0}
            self.fund_score[dt] = fund_score
            self.fund_score_raw[dt] = fund_score_raw

        self.asset_market_price_history[dt] = AssetValue(prices=asset_cur_price, positions=asset_cur_position).sum() 
        asset_w = AssetValue(prices=asset_cur_price, positions=asset_cur_position).get_weight().__dict__
        self.asset_weights_history[dt] = { index_id : w_i for index_id, w_i in asset_w.items() if w_i > 0}
        if traded_list is not None:
            if len(traded_list) > 0:
                self.trade_history[dt] = traded_list
        self.target_allocation[dt] = target_allocation

    def _calc_stat(self, df):
        year = df.shape[0]/self.TRADING_DAYS_PER_YEAR
        total_return = df['mv'][-1] / df['mv'][0]
        try:
            five_year_return = (df['mv'][-1] / df['mv'][-(5 * self.TRADING_DAYS_PER_YEAR)] - 1)
        except:
            five_year_return = five_year_return = (df['mv'][-1] / df['mv'][0] - 1)
        annualized_return = np.exp(np.log(total_return)/year) - 1
        annualized_volatiltity = (df['mv'].shift(1) / df['mv']).std() * np.sqrt((df.shape[0] - 1) / year)
        sharpe = annualized_return / annualized_volatiltity
        mdd = 1 - (df.loc[:, 'mv'] / df.loc[:, 'mv'].rolling(10000, min_periods=1).max()).min()
        mdd_part1 = (df.loc[:, 'mv'] / df.loc[:, 'mv'].rolling(10000, min_periods=1).max())
        mdd_date1 = df.loc[:mdd_part1.idxmin(),'mv'].idxmax()
        mdd_date2 = mdd_part1.idxmin()
        w = copy.deepcopy(self.saa_weight)
        w['mdd'] = mdd
        w['annual_ret'] = annualized_return
        w['sharpe'] = sharpe
        w['5_year_ret'] = five_year_return
        w['annual_vol'] = annualized_volatiltity
        w['mdd_d1'] = mdd_date1
        w['mdd_d2'] = mdd_date2
        w['market_value'] = df
        return w

    def get_asset_stat(self):
        self.asset_mv = pd.DataFrame([ {'date':k, 'mv':v} for k,v in self.asset_market_price_history.items()]).set_index('date')
        w = self._calc_stat(self.asset_mv.copy())
        return w
    
    def get_fund_stat(self):
        self.fund_mv = pd.DataFrame([ {'date':k, 'mv':v} for k,v in self.fund_marktet_price_history.items()]).set_index('date')
        w = self._calc_stat(self.fund_mv.copy())
        return w

    def get_fund_trade(self):
        fsdf = self.fund_score.copy()
        fund_info_df = self.fund_info.copy().set_index('fund_id')
        fund_trade = []
        for d in self.trade_history:
            f_t = [ i.__dict__  for i in self.trade_history[d] if isinstance(i, FundTrade)]
            fund_trade.extend(f_t)
        if len(fund_trade) < 1:
            return pd.DataFrame()
        ft_res = []
        for ft in fund_trade:
            ft['desc_name'] = fund_info_df.loc[ft['fund_id'], 'desc_name']
            index_id = fund_info_df.loc[ft['fund_id'], 'index_id']
            try:
                s = fsdf[ft['submit_date']][index_id][ft['fund_id']]
            except:
                s = np.nan
            ft['submit_d_score'] = s
            fund_id = ft['fund_id']
            submit_d = ft['submit_date']
            traded_d = ft['trade_date']
            ft['before_w']  = self.fund_weights_history[submit_d].get(fund_id,0)
            ft['after_w'] = self.fund_weights_history[traded_d].get(fund_id,0)
            ft_res.append(ft)
        fund_trade_df = pd.DataFrame(ft_res)
        return fund_trade_df

    def get_asset_trade(self):
        asset_trade = []
        for d in self.trade_history:
            a_t = [ i.__dict__  for i in self.trade_history[d] if not isinstance(i, FundTrade)]
            asset_trade.extend(a_t) 
        res = []
        for dic in asset_trade: 
            index_id = dic['index_id']
            submit_d = dic['submit_date']
            traded_d = dic['trade_date']
            dic['before_w'] = self.asset_weights_history[submit_d].get(index_id, 0)
            dic['after_w'] = self.asset_weights_history[traded_d].get(index_id, 0)
            res.append(dic)
        return pd.DataFrame(res)

    def backtest_asset_plot(self):
        bt_type = 'asset'
        print(f'{bt_type} report')
        w = self.get_asset_stat()
        del w['market_value']
        pprint(w)
        AssetPainter.plot_asset_weights(self.asset_weights_history)
        HybridPainter.plot_market_value(self.asset_mv, bt_type, self.index_price, self.saa_weight)
        AssetPainter.plot_asset_mdd_period(self.asset_mv, self.saa_weight, self.index_price)

    def backtest_fund_plot(self):
        bt_type = 'fund'
        print(f'{bt_type} report')
        w = self.get_fund_stat()
        del w['market_value']
        pprint(w)
        
        FundPainter.plot_fund_weights(self.fund_weights_history, self.fund_cash_history, self.fund_marktet_price_history)
        HybridPainter.plot_market_value(self.fund_mv, bt_type, self.index_price, self.saa_weight)
        FundPainter.plot_fund_mdd_periods(self.fund_mv, self.fund_weights_history, self.fund_nav, self.fund_info)
        HybridPainter.plot_asset_fund_mv_diff(self.asset_mv, self.fund_mv)

    def plot_whole(self):
        bt_type = 'asset'
        print(f'{bt_type} report')
        w = self.get_asset_stat()
        del w['market_value']
        pprint(w)
        AssetPainter.plot_asset_weights(self.asset_weights_history)
        HybridPainter.plot_market_value(self.asset_mv, bt_type, self.index_price, self.saa_weight)
        AssetPainter.plot_asset_mdd_period(self.asset_mv, self.saa_weight, self.index_price)
        
        bt_type = 'fund'
        print(f'{bt_type} report')
        w = self.get_fund_stat()
        del w['market_value']
        pprint(w)
        FundPainter.plot_fund_weights(self.fund_weights_history, self.fund_cash_history, self.fund_marktet_price_history)
        HybridPainter.plot_market_value(self.fund_mv, bt_type, self.index_price, self.saa_weight)
        FundPainter.plot_fund_mdd_periods(self.fund_mv, self.fund_weights_history, self.fund_nav, self.fund_info)
        HybridPainter.plot_asset_fund_mv_diff(self.asset_mv, self.fund_mv)
    
    def _plot_fund_score(self, asset, is_tuning):
        FundPainter.plot_fund_score(self.fund_mv, 
                                    self.fund_weights_history, 
                                    self.trade_history,
                                    self.index_price, 
                                    self.asset_weights_history,
                                    self.fund_info,
                                    self.fund_nav,
                                    self.fund_score,
                                    self.fund_score_raw,
                                    self.fund_indicator,
                                    asset,
                                    is_tuning,
                                    )
   
    def _index_pct_plot(self, index_id:str, saa_mv:pd.DataFrame, taa_mv:pd.DataFrame):
        AssetPainter.plot_taa_analysis(saa_mv, taa_mv, index_id, self.index_pct, self.taa_params, self.index_price)
        
    def _plot_taa_saa(self, saa_mv, taa_mv, index_id):
        AssetPainter.plot_asset_taa_saa(saa_mv, taa_mv, index_id, self.index_pct)
       